@@ -1,5 +1,5 @@
 --[[
-Copyright (c) 2017, Vsevolod Stakhov <vsevolod@highsecure.ru>
+Copyright (c) 2020, Vsevolod Stakhov <vsevolod@highsecure.ru>
 
 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
@@ -76,7 +76,7 @@ local settings = {
   allow_hdrfrom_mismatch_sign_networks = false,
   allow_hdrfrom_multiple = false,
   allow_username_mismatch = false,
-  auth_only = true,
+  sign_authenticated = true,
   domain = {},
   path = string.format('%s/%s/%s', rspamd_paths['DBDIR'], 'arc', '$domain.$selector.key'),
   sign_local = true,
@@ -88,13 +88,13 @@ local settings = {
   use_redis = false,
   key_prefix = 'arc_keys', -- default hash name
   reuse_auth_results = false, -- Reuse the existing authentication results
+  whitelisted_signers_map = nil, -- Trusted signers domains
+  allowed_ids = nil, -- Allowed settings id
+  forbidden_ids = nil, -- Banned settings id
 }
 
 -- To match normal AR
-local ar_settings = {
-  add_smtp_user = true,
-  stop_chars = ';'
-}
+local ar_settings = auth_results.default_settings
 
 local function parse_arc_header(hdr, target)
   -- Split elements by ';' and trim spaces
@@ -121,20 +121,27 @@ local function parse_arc_header(hdr, target)
     target[i].header = hdr[i].decoded
     target[i].raw_header = hdr[i].value
   end
+
+  -- sort by i= attribute
+  table.sort(target, function(a, b)
+    return (a.i or 0) < (b.i or 0)
+  end)
 end
 
 local function arc_validate_seals(task, seals, sigs, seal_headers, sig_headers)
   for i = 1,#seals do
     if (sigs[i].i or 0) ~= i then
-      rspamd_logger.infox(task, 'bad i value for signature: %s, expected %s',
-        sigs[i].i, i)
-      task:insert_result(arc_symbols['invalid'], 1.0, 'invalid count of seals and signatures')
+      local fail_reason = string.format('bad i for signature: %d, expected %d; d=%s',
+          sigs[i].i, i, sigs[i].d)
+      rspamd_logger.infox(task, fail_reason)
+      task:insert_result(arc_symbols['invalid'], 1.0, fail_reason)
       return false
     end
     if (seals[i].i or 0) ~= i then
-      rspamd_logger.infox(task, 'bad i value for seal: %s, expected %s',
-        seals[i].i, i)
-      task:insert_result(arc_symbols['invalid'], 1.0, 'invalid count of seals and signatures')
+      local fail_reason = string.format('bad i for seal: %d, expected %d; d=%s',
+          seals[i].i, i, seals[i].d)
+      rspamd_logger.infox(task, fail_reason)
+      task:insert_result(arc_symbols['invalid'], 1.0,fail_reason)
       return false
     end
 
@@ -174,7 +181,7 @@ local function arc_callback(task)
     -- We mandate that count of seals is equal to count of signatures
     rspamd_logger.infox(task, 'number of seals (%s) is not equal to number of signatures (%s)',
         #arc_seal_headers, #arc_sig_headers)
-    task:insert_result(arc_symbols['invalid'], 'invalid count of seals and signatures')
+    task:insert_result(arc_symbols['invalid'], 1.0, 'invalid count of seals and signatures')
     return
   end
 
@@ -183,7 +190,8 @@ local function arc_callback(task)
     sigs = {},
     checked = 0,
     res = 'success',
-    errors = {}
+    errors = {},
+    allowed_by_trusted = false
   }
 
   parse_arc_header(arc_seal_headers, cbdata.seals)
@@ -217,26 +225,36 @@ local function arc_callback(task)
   task:cache_set('arc-sigs', cbdata.sigs)
   task:cache_set('arc-seals', cbdata.seals)
 
-  local function arc_seal_cb(_, res, err, domain)
-    cbdata.checked = cbdata.checked + 1
-    lua_util.debugm(N, task, 'checked arc seal: %s(%s), %s processed',
-        res, err, cbdata.checked)
+  local function gen_arc_seal_cb(sig)
+    return function (_, res, err, domain)
+      cbdata.checked = cbdata.checked + 1
+      lua_util.debugm(N, task, 'checked arc seal: %s(%s), %s processed',
+          res, err, cbdata.checked)
 
-    if not res then
-      cbdata.res = 'fail'
-      if err and domain then
-        table.insert(cbdata.errors, string.format('sig:%s:%s', domain, err))
+      if not res then
+        cbdata.res = 'fail'
+        if err and domain then
+          table.insert(cbdata.errors, string.format('sig:%s:%s', domain, err))
+        end
       end
-    end
 
-    if cbdata.checked == #arc_sig_headers then
-      if cbdata.res == 'success' then
-        task:insert_result(arc_symbols['allow'], 1.0, 'i=' ..
-            tostring(cbdata.checked))
-      else
-        task:insert_result(arc_symbols['reject'], 1.0,
-          rspamd_logger.slog('seal check failed: %s, %s', cbdata.res,
-            cbdata.errors))
+      if settings.whitelisted_signers_map and cbdata.res == 'success' then
+        if settings.whitelisted_signers_map:get_key(sig.d) then
+          -- Whitelisted signer has been found in a valid chain
+          task:insert_result(arc_symbols.trusted_allow, 1.0,
+              string.format('%s:s=%s:i=%d', domain, sig.s, cbdata.checked))
+        end
+      end
+
+      if cbdata.checked == #arc_sig_headers then
+        if cbdata.res == 'success' then
+          task:insert_result(arc_symbols.allow, 1.0, string.format('%s:s=%s:i=%d',
+              domain, sig.s, cbdata.checked))
+        else
+          task:insert_result(arc_symbols.reject, 1.0,
+              rspamd_logger.slog('seal check failed: %s, %s', cbdata.res,
+                  cbdata.errors))
+        end
       end
     end
   end
@@ -256,10 +274,11 @@ local function arc_callback(task)
       cbdata.checked = 0
       fun.each(
         function(sig)
-          local ret, lerr = dkim_verify(task, sig.header, arc_seal_cb, 'arc-seal')
+          local ret, lerr = dkim_verify(task, sig.header, gen_arc_seal_cb(sig), 'arc-seal')
           if not ret then
             cbdata.res = 'fail'
-            table.insert(cbdata.errors, string.format('sig:%s:%s', sig.d or '', lerr))
+            table.insert(cbdata.errors, string.format('seal:%s:s=%s:i=%s:%s',
+                sig.d or '', sig.s or '', sig.i or '', lerr))
             cbdata.checked = cbdata.checked + 1
             lua_util.debugm(N, task, 'checked arc seal %s: %s(%s), %s processed',
               sig.d, ret, lerr, cbdata.checked)
@@ -397,6 +416,24 @@ rspamd_config:register_symbol({
   groups = {'arc'},
 })
 
+if settings.whitelisted_signers_map then
+  local lua_maps = require "lua_maps"
+  settings.whitelisted_signers_map = lua_maps.map_add_from_ucl(settings.whitelisted_signers_map,
+      'set',
+      'ARC trusted signers domains')
+  if settings.whitelisted_signers_map then
+    arc_symbols.trusted_allow = arc_symbols.trusted_allow or 'ARC_ALLOW_TRUSTED'
+    rspamd_config:register_symbol({
+      name = arc_symbols.trusted_allow,
+      parent = id,
+      type = 'virtual',
+      score = -2.0,
+      group = 'policies',
+      groups = {'arc'},
+    })
+  end
+end
+
 rspamd_config:register_dependency('ARC_CALLBACK', symbols['spf_allow_symbol'])
 rspamd_config:register_dependency('ARC_CALLBACK', symbols['dkim_allow_symbol'])
 
@@ -509,46 +546,94 @@ local function arc_sign_seal(task, params, header)
         'ARC-Seal', cur_arc_seal) }
     }
   })
-  task:insert_result(settings.sign_symbol, 1.0, string.format('i=%d', cur_idx))
+  task:insert_result(settings.sign_symbol, 1.0,
+      string.format('%s:s=%s:i=%d', params.domain, params.selector, cur_idx))
 end
 
-local function do_sign(task, p)
+local function prepare_arc_selector(task, sel)
+  local arc_seals = task:cache_get('arc-seals')
+
+  sel.arc_cv = 'none'
+  sel.arc_idx = 1
+  sel.no_cache = true
+  sel.sign_type = 'arc-sign'
+
+  if arc_seals then
+    sel.arc_idx = #arc_seals + 1
+
+    local function default_arc_cv()
+      if task:has_symbol(arc_symbols.allow) then
+        sel.arc_cv = 'pass'
+      else
+        sel.arc_cv = 'fail'
+      end
+    end
+
+    if settings.reuse_auth_results then
+      local ar_header = task:get_header('Authentication-Results')
+
+      if ar_header then
+        local arc_match = string.match(ar_header, 'arc=(%w+)')
+
+        if arc_match then
+          if arc_match == 'none' or arc_match == 'pass' then
+            -- none should be converted to `pass`
+            sel.arc_cv = 'pass'
+          else
+            sel.arc_cv = 'fail'
+          end
+        else
+          default_arc_cv()
+        end
+      else
+        -- Cannot reuse, use normal path
+        default_arc_cv()
+      end
+    else
+      default_arc_cv()
+    end
+
+  end
+end
+
+local function do_sign(task, sign_params)
+  if sign_params.alg and sign_params.alg ~= 'rsa' then
+    -- No support for ed25519 keys
+    return
+  end
+
+  prepare_arc_selector(task, sign_params)
+
   if settings.check_pubkey then
-    local resolve_name = p.selector .. "._domainkey." .. p.domain
+    local resolve_name = sign_params.selector .. "._domainkey." .. sign_params.domain
     task:get_resolver():resolve_txt({
       task = task,
       name = resolve_name,
       callback = function(_, _, results, err)
         if not err and results and results[1] then
-          p.pubkey = results[1]
-          p.strict_pubkey_check = not settings.allow_pubkey_mismatch
+          sign_params.pubkey = results[1]
+          sign_params.strict_pubkey_check = not settings.allow_pubkey_mismatch
         elseif not settings.allow_pubkey_mismatch then
           rspamd_logger.errx('public key for domain %s/%s is not found: %s, skip signing',
-              p.domain, p.selector, err)
+              sign_params.domain, sign_params.selector, err)
           return
         else
           rspamd_logger.infox('public key for domain %s/%s is not found: %s',
-              p.domain, p.selector, err)
+              sign_params.domain, sign_params.selector, err)
         end
 
-        local dret, hdr = dkim_sign(task, p)
+        local dret, hdr = dkim_sign(task, sign_params)
         if dret then
-          local sret, _ = arc_sign_seal(task, p, hdr)
-          if sret then
-            task:insert_result(settings.sign_symbol, 1.0)
-          end
+          arc_sign_seal(task, sign_params, hdr)
         end
 
       end,
       forced = true
     })
   else
-    local dret, hdr = dkim_sign(task, p)
+    local dret, hdr = dkim_sign(task, sign_params)
     if dret then
-      local sret, _ = arc_sign_seal(task, p, hdr)
-      if sret then
-        task:insert_result(settings.sign_symbol, 1.0)
-      end
+      arc_sign_seal(task, sign_params, hdr)
     end
   end
 end
@@ -558,60 +643,40 @@ local function sign_error(task, msg)
 end
 
 local function arc_signing_cb(task)
-  local arc_seals = task:cache_get('arc-seals')
-
   local ret, selectors = dkim_sign_tools.prepare_dkim_signing(N, task, settings)
 
   if not ret then
     return
   end
 
-  -- TODO: support multiple signatures here
-  local p = selectors[1]
-
-  p.arc_cv = 'none'
-  p.arc_idx = 1
-  p.no_cache = true
-  p.sign_type = 'arc-sign'
-
-  if arc_seals then
-    p.arc_idx = #arc_seals + 1
-
-    if task:has_symbol(arc_symbols.allow) then
-      p.arc_cv = 'pass'
-    else
-      p.arc_cv = 'fail'
-    end
-  end
-
   if settings.use_redis then
     dkim_sign_tools.sign_using_redis(N, task, settings, selectors, do_sign, sign_error)
   else
     if selectors.vault then
       dkim_sign_tools.sign_using_vault(N, task, settings, selectors, do_sign, sign_error)
     else
-      if ((p.key or p.rawkey) and p.selector) then
-        if p.key then
-          p.key = lua_util.template(p.key, {
-            domain = p.domain,
-            selector = p.selector
+      -- TODO: no support for multiple sigs
+      local cur_selector = selectors[1]
+      prepare_arc_selector(task, cur_selector)
+      if ((cur_selector.key or cur_selector.rawkey) and cur_selector.selector) then
+        if cur_selector.key then
+          cur_selector.key = lua_util.template(cur_selector.key, {
+            domain = cur_selector.domain,
+            selector = cur_selector.selector
           })
 
-          local exists,err = rspamd_util.file_exists(p.key)
+          local exists,err = rspamd_util.file_exists(cur_selector.key)
           if not exists then
             if err and err == 'No such file or directory' then
-              lua_util.debugm(N, task, 'cannot read key from %s: %s', p.key, err)
+              lua_util.debugm(N, task, 'cannot read key from %s: %s', cur_selector.key, err)
             else
-              rspamd_logger.warnx(task, 'cannot read key from %s: %s', p.key, err)
+              rspamd_logger.warnx(task, 'cannot read key from %s: %s', cur_selector.key, err)
             end
             return false
           end
         end
 
-        local dret, hdr = dkim_sign(task, p)
-        if dret then
-          return arc_sign_seal(task, p, hdr)
-        end
+        do_sign(task, cur_selector)
       else
         rspamd_logger.infox(task, 'key path or dkim selector unconfigured; no signing')
         return false
@@ -646,14 +711,24 @@ if settings.use_redis then
         'but module is configured to load keys from redis, disable arc signing')
     return
   end
+
+  settings.redis_params = redis_params
 end
 
-rspamd_config:register_symbol({
+local sym_reg_tbl = {
   name = settings['sign_symbol'],
   callback = arc_signing_cb,
   groups = {"policies", "arc"},
   score = 0.0,
-})
+}
+if type(settings.allowed_ids) == 'table' then
+  sym_reg_tbl.allowed_ids = settings.allowed_ids
+end
+if type(settings.forbidden_ids) == 'table' then
+  sym_reg_tbl.forbidden_ids = settings.forbidden_ids
+end
 
--- Do not sign unless valid
+rspamd_config:register_symbol(sym_reg_tbl)
+
+-- Do not sign unless checked
 rspamd_config:register_dependency(settings['sign_symbol'], 'ARC_CALLBACK')
